import numpy as np
from config.core import config
from pipeline import price_pipe
from processing.data_manager import load_dataset, save_pipeline
from sklearn.model_selection import train_test_split


def run_training() -> None:
    """Train the model."""

    # read training data
    data = load_dataset(file_name=config.app_config.training_data_file)

    # divide train and test
    X_train, X_test, y_train, y_test = train_test_split(
        data[config.model_config.features],  # predictors
        data[config.model_config.target],
        test_size=config.model_config.test_size,
        # we are setting the random seed here
        # for reproducibility
        random_state=config.model_config.random_state,
    )
    y_train = np.log(y_train)

    # print(f"X_train: {X_train.shape}, y_train: {y_train.shape}")
    # print(f"Type X_train: {type(X_train)}")
    # print(f"Type y_train: {type(y_train)}")
    # print(f"Length of X_train: {len(X_train)}")
    # print(f"Length of y_train: {len(y_train)}")
    # print("X_train: ")
    # print("-" * 20)
    # print(X_train)
    # print("y_train: ")
    # print("-" * 20)
    # print(y_train)
    # fit model
    price_pipe.fit(X_train, y_train)
    # # ---Change
    # price_pipe = price_pipe[:-1].transform(X_train) #Exclude the last step (Lasso)
    # persist trained model
    save_pipeline(pipeline_to_persist=price_pipe)


if __name__ == "__main__":
    run_training()



from feature_engine.datetime import DatetimeFeatures

from regression_model.config.core import config


def preprocess_dataframe(dataframe, config=config):
    # Drop rows with missing values in the target column
    dataframe.dropna(subset=[config.model_config.target], axis=0, inplace=True)
    # dataframe.drop(dataframe.loc[dataframe[config.model_config.target].isnull()].index,
    #                axis = 0, inplace = True)
    # Drop specified columns
    dataframe.drop(config.model_config.drop_features_pp, axis=1, inplace=True)

    # Drop duplicates, keeping the first occurrence
    dataframe.drop_duplicates(keep='first', inplace=True)

    # Extract datetime features
    date_time_features = DatetimeFeatures(
        variables=config.model_config.date_features,
        features_to_extract=['year'],
        drop_original=True,
        missing_values='ignore'
    )
    dataframe = date_time_features.fit_transform(dataframe)

    return dataframe



from feature_engine.encoding import OrdinalEncoder, RareLabelEncoder
from feature_engine.imputation import (  # noqa
    AddMissingIndicator,
    CategoricalImputer,
    MeanMedianImputer,
)
from feature_engine.selection import DropFeatures
from feature_engine.transformation import LogTransformer
from feature_engine.wrappers import SklearnTransformerWrapper  # noqa
from sklearn.feature_selection import SelectFromModel
from sklearn.linear_model import Lasso
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import Binarizer, MinMaxScaler  # noqa

from regression_model.config.core import config
from regression_model.processing import sklearn_preprocessors as skpp

price_pipe = Pipeline(
    [
        # ===== IMPUTATION =====
        # add missing indicator
        (
            "missing_indicator",
            AddMissingIndicator(variables=config.model_config.numerical_vars_with_na),
        ),
        # impute numerical variables with the mean
        (
            "mean_imputation",
            MeanMedianImputer(
                imputation_method="mean",
                variables=config.model_config.numerical_vars_with_na,
            ),
        ),
        # ===== TEMPORAL VARIABLES =====
        (
            "elapsed_time",
            skpp.ElapsedYearsTransformer(
                vars=config.model_config.temporal_vars,
                # ref_var = config.model_config.ref_var,
            ),
        ),
        # ===== DROP FEATURES =====
        ("drop_features", DropFeatures(features_to_drop=config.model_config.drop_features)),
        # # impute categorical variables with string missing
        # (
        #     "missing_imputation",
        #     CategoricalImputer(
        #         imputation_method="missing",
        #         variables=config.model_config.categorical_vars_with_na_missing,
        #     ),
        # ),
        # # impute categorical variables with string missing
        # (
        #     "frequent_imputation",
        #     CategoricalImputer(
        #         imputation_method="frequent",
        #         variables=config.model_config.categorical_vars_with_na_frequent,
        #     ),
        # ),
        # ==== DATA TYPES CONVERSION =====
        (
            'convert_to_int64',
            skpp.DataTypeConverter(
                vars=config.model_config.float_cols,
                data_type=config.model_config.to_int64
            )
        ),
        # ==== VARIABLE TRANSFORMATION =====
        (
            "log",
            LogTransformer(
                    variables=config.model_config.numericals_log_vars
            )
        ),
        # ('yeojohnson', YeoJohnsonTransformer(variables=config.model_config.numericals_log_vars)),
        # #Binarizer
        # (
        #     "binarizer",
        #     SklearnTransformerWrapper(
        #         transformer=Binarizer(threshold=0),
        #         variables=config.model_config.binarize_vars,
        #     ),
        # ),
        # ==== mappers ====
        (
            "mapper_type_vars",
            skpp.Mapper(
                variables=config.model_config.type_vars,
                mappings=config.model_config.type_vars_mappings,
            ),
        ),
        # ==== CATEGORICAL ENCODING ====
        (
            "rare_label_encoder",
            RareLabelEncoder(
                tol=config.model_config.rare_labels_tolerance, n_categories=1,
                variables=config.model_config.categorical_vars
            ),
        ),
        # encode categorical variables using the target mean
        (
            "categorical_encoder",
            OrdinalEncoder(
                encoding_method="ordered",
                variables=config.model_config.categorical_vars,
            ),
        ),
        # ==== MinMaxScaler ====
        ("scaler", MinMaxScaler()),
        # ==== Model Selection ====
        (
            'selector',
            SelectFromModel(
                Lasso(
                    alpha=config.model_config.alpha,
                    random_state=config.model_config.random_state,
                    )
            )
        ),
        (
            "Lasso",
            Lasso(
                alpha=config.model_config.alpha,
                random_state=config.model_config.random_state,
            ),
        ),
    ]
)

# print(price_pipe[2:])



from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from pydantic import BaseModel, ValidationError

from regression_model.config.core import config
from regression_model.processing.datapreprocessor import preprocess_dataframe  # noqa


def drop_na_inputs(*, input_data: pd.DataFrame) -> pd.DataFrame:
    """Check model inputs for na values and filter."""
    validated_data = input_data.copy()
    # new_vars_with_na = [
    #     var
    #     for var in config.model_config.features
    #     if var
    #     # not in config.model_config.categorical_vars_with_na_frequent
    #     # + config.model_config.categorical_vars_with_na_missing
    #     + config.model_config.numerical_vars_with_na
    #     and validated_data[var].isnull().sum() > 0
    # ]
    new_vars_with_na = [var
                        for var in config.model_config.features
                        if validated_data[var].isnull().sum() > 0]
    validated_data.dropna(subset=new_vars_with_na, inplace=True)
    return validated_data


def validate_inputs(*, input_data: pd.DataFrame) -> Tuple[pd.DataFrame, Optional[dict]]:
    """Check model inputs for unprocessable values."""
    # input_data = pd.DataFrame(input_data, parse_dates=config.model_config.date_features)
    # input_data = preprocess_dataframe(input_data)
    # drop target
    relevant_data = input_data[config.model_config.features].copy()
    # By Passing for now
    # validated_data = drop_na_inputs(input_data=relevant_data)
    validated_data = relevant_data
    errors = None

    try:
        # replace numpy nans so that pydantic can validate
        MultipleDataInputs(
            inputs=validated_data.replace({np.nan: None}).to_dict(orient="records")
        )
    except ValidationError as error:
        errors = error.json()

    return validated_data, errors


class DataInputSchema(BaseModel):
    Location1: Optional[str]
    Type: Optional[str]
    Bedrooms: Optional[float]
    Bathrooms: Optional[float]
    Size_in_SqYds: Optional[float]
    Price_in_millions: Optional[float]
    Parking_Spaces: Optional[int]
    Floors_in_Building: Optional[float]
    Elevators: Optional[float]
    Lobby_in_Building: Optional[int]
    Double_Glazed_Windows: Optional[int]
    Central_Air_Conditioning: Optional[int]
    Central_Heating: Optional[int]
    Waste_Disposal: Optional[int]
    Furnished: Optional[int]
    Service_Elevators_in_Building: Optional[int]
    Flooring: Optional[int]
    Electricity_Backup: Optional[int]
    Servant_Quarters: Optional[int]
    Study_Room: Optional[int]
    Prayer_Room: Optional[int]
    Powder_Room: Optional[int]
    Gym: Optional[int]
    Lounge_or_Sitting_Room: Optional[int]
    Laundry_Room: Optional[int]
    Business_Center_or_Media_Room_in_Building: Optional[int]
    Satellite_or_Cable_TV_Ready: Optional[int]
    Broadband_Internet_Access: Optional[int]
    Intercom: Optional[int]
    Conference_Room_in_Building: Optional[int]
    Community_Swimming_Pool: Optional[int]
    Community_Lawn_or_Garden: Optional[int]
    Community_Gym: Optional[int]
    Community_Center: Optional[int]
    First_Aid_or_Medical_Centre: Optional[int]
    Day_Care_center: Optional[int]
    Kids_Play_Area: Optional[int]
    Mosque: Optional[int]
    Barbeque_Area: Optional[int]
    Lawn_or_Garden: Optional[int]
    Swimming_Pool: Optional[int]
    Sauna: Optional[int]
    Jacuzzi: Optional[int]
    Nearby_Schools: Optional[int]
    Nearby_Hospital: Optional[int]
    Nearby_Shopping_Malls: Optional[int]
    Nearby_Restaurants: Optional[int]
    Nearby_Public_Transport_Service: Optional[int]
    Other_Nearby_Places: Optional[int]
    Security_Staff: Optional[int]
    Maintainance_Staff: Optional[int]
    Laundry_or_Dry_Cleaning_Facility: Optional[int]
    Facilities_for_Disabled: Optional[int]
    # this one is only to calculate temporal variable:
    Built_in_year_year: Optional[float]


class MultipleDataInputs(BaseModel):
    inputs: List[DataInputSchema]



import pytest

from regression_model.config.core import config
from regression_model.processing.sklearn_preprocessors import ElapsedYearsTransformer


def test_ElapsedYearsTransformer(sample_input_data):
    # Given
    transformer = ElapsedYearsTransformer(
        vars=config.model_config.temporal_vars,  # Built_in_year_year
        # reference_variable=config.model_config.ref_var,
    )
    assert sample_input_data["Built_in_year_year"].iat[3] == 2020.0

    # When
    subject = transformer.fit_transform(sample_input_data)

    # Then
    assert subject["Built_in_year_year"].iat[3] == 2020.0


def test_ElapsedYearsTransformer_invalid_variable(sample_input_data):
    # Given
    transformer = ElapsedYearsTransformer(
        vars=["InvalidVariable"],  # Invalid variable name
        # reference_variable=config.model_config.ref_var,
    )

    # When/Then
    # An exception should be raised when fitting the transformer
    with pytest.raises(KeyError):
        transformer.fit_transform(sample_input_data)


def test_ElapsedYearsTransformer_missing_variable(sample_input_data):
    # Given
    transformer = ElapsedYearsTransformer(
        vars=["MissingVariable"],  # Variable not present in the dataset
        # reference_variable = config.model_config.ref_var,
    )

    # When/Then
    # An exception should be raised when fitting the transformer
    with pytest.raises(KeyError):
        transformer.fit_transform(sample_input_data)



import math

import numpy as np

from regression_model.predict import make_prediction


def test_make_prediction(sample_input_data):
    # Given
    expected_first_prediction_value = 125
    expected_no_predictions = len(sample_input_data)
    # #print("Sample input data: ", sample_input_data)
    # print("Length of sample input data: ", len(sample_input_data))

    # When
    result = make_prediction(input_data=sample_input_data)

    # Then
    predictions = result.get("predictions")
    # print("Length of predictions: ", len(predictions))
    # print("Predictions: ",predictions)
    assert isinstance(predictions, list)
    assert isinstance(predictions[0], np.float64)
    assert result.get("errors") is None
    assert len(predictions) == expected_no_predictions
    assert math.isclose(predictions[0], expected_first_prediction_value, abs_tol=100)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path

from setuptools import find_packages, setup

# Package meta-data.
NAME = 'Lasso Regression Model'
DESCRIPTION = "Lasso Regression model package"
URL = "https://github.com/mm-mazhar/Deploying-ML-Lasso-Regression-Model"
EMAIL = "mazqoty.01@mail.com"
AUTHOR = "Mazhar"
REQUIRES_PYTHON = ">=3.6.0"


# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the
# Trove Classifier for that!
long_description = DESCRIPTION

# Load the package's VERSION file as a dictionary.
about = {}
ROOT_DIR = Path(__file__).resolve().parent
REQUIREMENTS_DIR = ROOT_DIR / 'requirements'
PACKAGE_DIR = ROOT_DIR / 'regression_model'
with open(PACKAGE_DIR / "VERSION") as f:
    _version = f.read().strip()
    about["__version__"] = _version


# What packages are required for this module to be executed?
def list_reqs(fname="requirements.txt"):
    with open(REQUIREMENTS_DIR / fname) as fd:
        return fd.read().splitlines()

# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=("tests",)),
    package_data={"regression_model": ["VERSION"]},
    install_requires=list_reqs(),
    extras_require={},
    include_package_data=True,
    license="BSD-3",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)


from datetime import datetime
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class ElapsedYearsTransformer(BaseEstimator, TransformerMixin):
    """Calculate Elapsed Years From Reference Years"""
    def __init__(self, vars: List[str], ref_var: Optional[str] = None):
        self.vars = None
        self.ref_var = None
        if not isinstance(vars, list):
            raise ValueError(f'{vars} must be a list')
        if ref_var is None:
            pass
        else:
            if not isinstance(ref_var, str):  # ref_var is a column name in the dataframe
                raise ValueError(f'{ref_var} must be a string')
        self.vars = vars
        self.ref_var = ref_var
        # print("Type of self.ref_var in init: ", type(self.ref_var))
        # print("Value of self.ref_var in init: ", self.ref_var)

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        # we need this step to fit the sklearn pipeline
        return self

    def transform(self, X: pd.DataFrame, y: pd.Series = None) -> Union[pd.DataFrame, None]:
        # so that we do not over-write the original dataframe
        X = X.copy()
        if not isinstance(self.ref_var, str):
            self.ref_var = datetime.now().year  # type: ignore
            X[str(self.ref_var)] = self.ref_var
            for var in self.vars:  # type: ignore
                X['diff_from_' + str(self.ref_var) + '-' + var] = X[str(self.ref_var)] - X[var]
            X.drop(str(self.ref_var), axis=1, inplace=True)
            # print("ElapsedYearsTransformer: Executed if")
            return X
        else:
            if self.ref_var in X.columns:
                for var in self.vars:  # type: ignore
                    X['diff_from_' + str(self.ref_var) + '-' + var] = X[str(self.ref_var)] - X[var]
                return X
            else:
                print(f'{self.ref_var} is not in dataframe columns: {X.columns.values}')
                # print("ElapsedYearsTransformer: Executed else else")
                return None


class DataTypeConverter(BaseEstimator, TransformerMixin):
    """ Custom Transformer to convert columns to int32 """
    def __init__(self, vars: List[str], data_type: str):
        self.vars = None
        self.data_type = None
        if not isinstance(vars, list):
            raise ValueError(f'{vars} must be a list')
        if not isinstance(data_type, str):  # ref_var is a column name in the dataframe
            raise ValueError(f'{data_type} must be a string')
        self.vars = vars
        self.data_type = data_type

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        for var in self.vars:  # type: ignore
            X[var] = X[var].astype(self.data_type)
        return X


class Mapper(BaseEstimator, TransformerMixin):
    """String to numbers Mapper."""
    def __init__(self, variables: List[str], mappings: dict):  # vars --> list of cols to be mapped
        self.variables = None
        self.mappings = None
        if not isinstance(variables, list):
            raise ValueError(f'{variables} must be a list')
        self.variables = variables
        self.mappings = mappings

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        # we need the fit statement to accomodate the sklearn pipeline
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        # so that we do not over-write the original dataframe
        X = X.copy()
        for feature in self.variables:  # type: ignore
            X[feature] = X[feature].map(self.mappings)

        return X


class MeanImputer(BaseEstimator, TransformerMixin):
    """Numerical missing value imputer (with Mean)."""
    def __init__(self, variables: List[str]):
        if not isinstance(variables, list):
            raise ValueError('variables should be a list')
        self.variables = variables
        self.imputer_dict_ = {}  # type: ignore

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        # persist mean values in a dictionary
        # self.imputer_dict_ = {}
        self.imputer_dict_ = X[self.variables].mean().to_dict()
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.copy()
        for feature in self.variables:
            X[feature].fillna(self.imputer_dict_[feature], inplace=True)
        return X


class ModeImputer(BaseEstimator, TransformerMixin):
    """Numerical missing value imputer (with Mode)."""

    def __init__(self, variables: List[str]):
        if not isinstance(variables, list):
            raise ValueError('variables should be a list')
        self.variables = variables
        self.imputer_dict_ = {}  # type: ignore

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        # persist mean values in a dictionary
        # self.imputer_dict_ = {}
        self.imputer_dict_ = X[self.variables].mode().to_dict()
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.copy()
        for feature in self.variables:
            X[feature].fillna(self.imputer_dict_[feature], inplace=True)
        return X


class RareLabelCategoricalEncoder(BaseEstimator, TransformerMixin):
    """Groups infrequent categories into a single string"""

    def __init__(self, variables: List[str],  tol: float = 0.01):

        if not isinstance(variables, list):
            raise ValueError('variables should be a list')
        self.tol = tol
        self.variables = variables

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        # persist frequent labels in dictionary
        self.encoder_dict_ = {}

        for var in self.variables:
            # the encoder will learn the most frequent categories
            t = pd.Series(X[var]).value_counts(normalize=True)
            # frequent labels:
            self.encoder_dict_[var] = list(t[t >= self.tol].index)

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.copy()
        for feature in self.variables:
            X[feature] = np.where(
                X[feature].isin(self.encoder_dict_[feature]), X[feature], "Rare")

        return X


class CategoricalEncoder(BaseEstimator, TransformerMixin):
    """String to numbers categorical encoder."""
    def __init__(self, variables: List[str]):

        if not isinstance(variables, list):
            raise ValueError('variables should be a list')
        self.variables = variables

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        temp = pd.concat([X, y], axis=1)
        temp.columns = list(X.columns) + ["target"]

        # persist transforming dictionary
        self.encoder_dict_ = {}

        for var in self.variables:
            t = temp.groupby([var])["target"].mean().sort_values(ascending=True).index
            self.encoder_dict_[var] = {k: i for i, k in enumerate(t, 0)}

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        # encode labels
        X = X.copy()
        for feature in self.variables:
            X[feature] = X[feature].map(self.encoder_dict_[feature])

        return X



import typing as t

import numpy as np
import pandas as pd

from regression_model import __version__ as _version
from regression_model.config.core import config
from regression_model.processing.data_manager import load_pipeline
from regression_model.processing.validation import validate_inputs

pipeline_file_name = f"{config.app_config.pipeline_save_file}{_version}.pkl"
_price_pipe = load_pipeline(file_name=pipeline_file_name)


def make_prediction(*, input_data: t.Union[pd.DataFrame, dict],) -> dict:
    """Make a prediction using a saved model pipeline."""
    data = pd.DataFrame(input_data)
    validated_data, errors = validate_inputs(input_data=data)
    results = {"predictions": None, "version": _version, "errors": errors}

    # print("Validated data: ", validated_data)

    if not errors:
        transformed_data = _price_pipe[:-1].transform(
            X=validated_data[config.model_config.features]
            )
        predictions = _price_pipe[-1].predict(transformed_data)
        # predictions = _price_pipe.predict(X=validated_data[config.model_config.features])
        results = {
            "predictions": [np.exp(pred) for pred in predictions],  # type: ignore
            "version": _version,
            "errors": errors,
        }
    return results
